import clodius.fpark as fp


def test_map():
    a = fp.FakeSparkContext.parallelize([1, 2, 3, 4])

    b = a.map(lambda x: x + 1)
    assert(b.collect()[0] == 2)


def test_group_by_key():
    a = fp.FakeSparkContext.parallelize(
        [(1, 2), (1, 3), (1, 4), (2, 5), (2, 6)])

    a.groupByKey()
    # TODO: Make assertions about result


def test_textFile():
    a = fp.FakeSparkContext.textFile(
        'test/sample_data/piecewise-file/part-00000')
    parts = a.collect()
    assert(len(parts) == 2)

    a = fp.FakeSparkContext.textFile('test/sample_data/piecewise-file/')
    parts = a.collect()
    assert(len(parts) == 4)



from clodius.tiles.format import format_dense_tile
from clodius.tiles.utils import tile_bounds
import pandas as pd

import numpy as np
import h5py


def csv_to_points(csv_file, output_file, max_zoom: int = 30):
    '''
    Convert a csv file containing points to a numpy array
    of [[x,y]] values.

    Parameters:
    -----------
    csv_file: string
        The filename of the data file

    '''
    df = pd.read_csv(csv_file)

    min_x = df['x'].min()
    max_x = df['x'].max()
    min_y = df['y'].min()
    max_y = df['y'].max()

    width = max_x - min_x
    height = max_y - min_y
    max_width = max(width, height)

    with h5py.File(output_file, 'w') as f_out:
        dataset = f_out.create_dataset(
            'values', (len(df), 2), compression='gzip', dtype=np.float32)
        dataset[:] = df.reindex(columns=['x', 'y']).values

        dataset.attrs['min_x'] = min_x
        dataset.attrs['max_x'] = max_x
        dataset.attrs['min_y'] = min_y
        dataset.attrs['max_y'] = max_y
        dataset.attrs['max_zoom'] = max_zoom
        dataset.attrs['max_width'] = max_width

    return df.reindex(columns=['x', 'y'])


def tileset_info(points_file):
    '''
    Calculate the extent, etc...
    '''
    with h5py.File(points_file, 'r') as f_in:
        attrs = f_in['values'].attrs

        return {
            'min_pos': [float(attrs['min_x']), float(attrs['min_y'])],
            'max_pos': [float(attrs['max_x']), float(attrs['max_y'])],
            'max_width': float(attrs['max_width']),
            'max_zoom': int(attrs['max_zoom']),
            'mirror_tiles': 'false'
        }


def filter_points(data, extent):
    '''
    Filter points that are within the extent

    Parameters:
    -----------
    data: [[]]
        A 2D numpy array containing x,y values

    extent: [x_start, x_end, y_start, y_end]
        The region we want to return points within

    Returns
    -------
    data: [[]]
        A 2D numpy array containing x,y values
    '''
    # print("extent:", extent)
    # print("data.shape", data.shape, data[:,0])
    data = data[data[:, 0] > extent[0]]
    data = data[data[:, 0] < extent[2]]

    data = data[data[:, 1] > extent[1]]
    data = data[data[:, 1] < extent[3]]

    return data


def density_tiles(points_file, z, x, y, width=1, height=1):
    '''
    Get a 2D histogram of the given region. If the height and
    width are specified, then we need to partition this into
    multiple returned tiles.
    '''
    returns = []

    with h5py.File(points_file, 'r') as f:
        # get all the points in the region
        all_points = filter_points(
            f['values'][:],
            tile_bounds(tileset_info(points_file), z, x, y, width, height)
        )

        for i in range(width):
            for j in range(height):
                # filter from the larger subregion
                filtered_points = filter_points(
                    all_points,
                    tile_bounds(tileset_info(points_file), z, x + i, y + j)
                )

                dt = np.histogram2d(filtered_points[:, 0],
                                    filtered_points[:, 1], bins=256)[0].T
                dt[dt == 0.] = np.nan

                returns += [((z, x + i, y + j), dt)]

        return returns


def tiles(points_file, z, x, y, width=1, height=1):
    return [
        (tile_pos, format_dense_tile(data.flatten())) for
        (tile_pos, data) in density_tiles(points_file, z, x, y, width, height)
    ]



import h5py
import logging

logger = logging.getLogger(__name__)


def get_cooler_info(file_path):
    """Get information of a cooler file.

    Args:
        file_path (str): Path to a cooler file.

    Returns:
        dict: Dictionary containing basic information about the cooler file.
    """
    TILE_SIZE = 256
    CHROM_CUM_LEN = 0
    with h5py.File(file_path, 'r') as f:
        max_zoom = f.attrs.get('max-zoom')

        if max_zoom is None:
            logger.info('no zoom found')
            raise ValueError(
                'The `max_zoom` attribute is missing.'
            )

        total_length = CHROM_CUM_LEN
        max_zoom = f.attrs['max-zoom']
        bin_size = int(f[str(max_zoom)].attrs['bin-size'])

        max_width = bin_size * TILE_SIZE * 2**max_zoom

        info = {
            'min_pos': [0.0, 0.0],
            'max_pos': [total_length, total_length],
            'max_zoom': max_zoom,
            'max_width': max_width,
            'bins_per_dimension': TILE_SIZE,
        }

        return info



from __future__ import print_function

import clodius.db_tiles as cdt


def test_get_tileset_info():
    filename = 'test/sample_data/arrowhead_domains_short.txt.multires.db'
    t = cdt.get_tileset_info(filename)

    assert(t['zoom_step'] == 1)
    assert(t['max_length'] == 3095693981)
    assert(t['max_width'] > 4000000000)
    assert(t['max_width'] < 5000000000)


def test_get_tiles():
    filename = 'test/sample_data/arrowhead_domains_short.txt.multires.db'

    cdt.get_2d_tiles(filename, 0, 0, 0, numx=1, numy=1)
    # TODO: Make an assertion



import os
import shortuuid
import time

import_time = time.time()


def describe_dataset(args_raw, args_parsed):
    '''
    Create a JSON structure describing how this dataset
    was created. The input is the parsed set of command line arguments used to
    run this program.

    :param args: The arguments used to run the program.
    :return: A JSON file with the arguments as well as some other identifying information.
    '''
    return {'uuid': shortuuid.uuid(),
            'cwd': os.getcwd(),
            'args_raw': args_raw,
            'command': " ".join(args_raw),
            'args_parsed': vars(args_parsed),
            'import_time': time.strftime("%Y-%m-%d %H:%M:%S %z", time.localtime(import_time)),
            'current_time': time.strftime("%Y-%m-%d %H:%M:%S %z"),
            'elapsed_time': round(time.time() - import_time, 2)}



import clodius.tiles.bigwig as hgbi
import os.path as op
import numpy as np
import base64


def arr_eq_nan(a, b):
    return ((a == b) | (np.isnan(a) & np.isnan(b))).all()


def test_bigwig_tiles():
    filename = op.join(
        'data',
        'wgEncodeCaltechRnaSeqHuvecR1x75dTh1014IlnaPlusSignalRep2.bigWig'
    )

    mean_tile = hgbi.tiles(filename, ['x.0.0'])
    mean_mean_tile = hgbi.tiles(filename, ['x.0.0.mean'])
    min_tile = hgbi.tiles(filename, ['x.0.0.min'])
    max_tile = hgbi.tiles(filename, ['x.0.0.max'])
    std_tile = hgbi.tiles(filename, ['x.0.0.std'])

    assert mean_tile[0][1]['max_value'] == mean_mean_tile[0][1]['max_value']
    assert mean_tile[0][1]['max_value'] > min_tile[0][1]['max_value']
    assert max_tile[0][1]['max_value'] > mean_tile[0][1]['max_value']
    assert max_tile[0][1]['max_value'] > mean_tile[0][1]['max_value'] + \
        std_tile[0][1]['max_value']

    min_max_tile = hgbi.tiles(filename, ['x.0.0.minMax'])
    whisker_tile = hgbi.tiles(filename, ['x.0.0.whisker'])

    mean_val = np.frombuffer(
        base64.b64decode(mean_tile[0][1]['dense']),
        dtype=mean_tile[0][1]['dtype']
    )

    min_val = np.frombuffer(
        base64.b64decode(min_tile[0][1]['dense']),
        dtype=min_tile[0][1]['dtype']
    )

    max_val = np.frombuffer(
        base64.b64decode(max_tile[0][1]['dense']),
        dtype=max_tile[0][1]['dtype']
    )

    std_val = np.frombuffer(
        base64.b64decode(std_tile[0][1]['dense']),
        dtype=std_tile[0][1]['dtype']
    )

    min_max_val = np.frombuffer(
        base64.b64decode(min_max_tile[0][1]['dense']),
        dtype=min_max_tile[0][1]['dtype']
    )

    whisker_val = np.frombuffer(
        base64.b64decode(whisker_tile[0][1]['dense']),
        dtype=whisker_tile[0][1]['dtype']
    )

    assert min_max_val.shape[0] == 2 * mean_val.shape[0]
    assert arr_eq_nan(min_max_val[::2], min_val)
    assert arr_eq_nan(min_max_val[1::2], max_val)

    assert whisker_val.shape[0] == 4 * mean_val.shape[0]
    assert arr_eq_nan(whisker_val[::4], min_val)
    assert arr_eq_nan(whisker_val[1::4], max_val)
    assert arr_eq_nan(whisker_val[2::4], mean_val)
    assert arr_eq_nan(whisker_val[3::4], std_val)


def test_tileset_info():
    filename = op.join(
        'data',
        'wgEncodeCaltechRnaSeqHuvecR1x75dTh1014IlnaPlusSignalRep2.bigWig'
    )

    tileset_info = hgbi.tileset_info(filename)

    assert len(tileset_info['aggregation_modes']) == 4
    assert tileset_info['aggregation_modes']['mean']
    assert tileset_info['aggregation_modes']['min']
    assert tileset_info['aggregation_modes']['max']
    assert tileset_info['aggregation_modes']['std']

    assert len(tileset_info['range_modes']) == 2
    assert tileset_info['range_modes']['minMax']
    assert tileset_info['range_modes']['whisker']


def test_natsorted():
    chromname_tests = [
        {
            'input': ['2', '3', '4', 'm', 'x', '1', 'y'],
            'expected': ['1', '2', '3', '4', 'x', 'y', 'm']
        },
        {
            'input': ['chr1', 'chr4', 'chr5', 'chr2',
                      'chr3', 'chrMT', 'chrY', 'chrX'],
            'expected': ['chr1', 'chr2', 'chr3', 'chr4',
                         'chr5', 'chrX', 'chrY', 'chrMT']
        }
    ]

    for test in chromname_tests:
        sorted_output = hgbi.natsorted(test['input'])
        assert sorted_output == test['expected'],\
            'Sorted output was %s\nExpected: %s' \
            % (sorted_output, test['expected'])



import clodius.tiles.files as hgfi


def test_infer_filetype():
    assert(hgfi.infer_filetype('matrix.mcool') == 'cooler')
    assert(hgfi.infer_filetype('matrix.txt') is None)
    assert(hgfi.infer_filetype('a.bigwig') == 'bigwig')
    assert(hgfi.infer_filetype('a.bigWig') == 'bigwig')


def test_infer_datatype():
    assert(hgfi.infer_datatype('cooler') == 'matrix')
    assert(hgfi.infer_datatype('bigwig') == 'vector')



from __future__ import print_function

import click.testing as clt
import clodius.cli.aggregate as cca
import clodius.db_tiles as cdt
import os
import os.path as op
import sqlite3
import tempfile

testdir = op.realpath(op.dirname(__file__))


def test_nonstandard_chrom():
    filename = 'test/sample_data/test_non_standard_chrom.bed'
    f = tempfile.NamedTemporaryFile(delete=False)

    ret = cca._bedfile(filename, f.name,
                       'hg19', None, False,
                       None, 100, 1024, None, None, 0)

    assert ret is None

    ret = cca._bedfile(filename, f.name,
                       'dfsdfs', None, False,
                       None, 100, 1024, None, None, 0)

    assert ret is None


def test_get_tileset_info():
    filename = 'test/sample_data/gene_annotations.short.db'
    t = cdt.get_tileset_info(filename)

    assert(t['zoom_step'] == 1)
    assert(t['max_length'] == 3137161264)
    assert(t['max_width'] > 4000000000)
    assert(t['max_width'] < 5000000000)


def test_table_created():
    check_table('test/sample_data/gene_annotations.short.db')


def check_table(filename):
    conn = sqlite3.connect(filename)
    c = conn.cursor()

    print("fetching...")

    '''
    for row in c.execute('SELECT * from intervals'):
        print ("row:", row)
    '''

    rows = c.execute(
        'SELECT * from intervals,position_index '
        'where intervals.id=position_index.id '
        'and zoomLevel < 1 '
        'and rStartPos > 2400000000 '
        'and rEndPos < 2500000000')
    counter = 0
    for row in rows:
        assert(row[3] > 2400000000)
        assert(row[4] < 2500000000)
        counter += 1

    assert(counter > 0)


def test_get_tiles():
    filename = 'test/sample_data/gene_annotations.short.db'

    cdt.get_tiles(filename, 18, 169283)[169283]
    # TODO: Make assertions about result
    # print("tiles:", tiles)
    # x = int(tiles[0]['xStart'])
    # fields = tiles[0]['fields']
    # TODO: Make assertions


def test_gene_annotations():
    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data', 'exon_unions_mm10.bed')
    f = tempfile.NamedTemporaryFile(delete=False)

    result = runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', '20', '--importance-column', '5',
         '--delimiter', '\t',
         '--assembly', 'mm10', '--output-file', f.name])

    # import traceback
    a, b, tb = result.exc_info
    '''
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    '''

    rows = cdt.get_tiles(f.name, 0, 0)
    assert(len(rows[0]) == 2)

    rows = cdt.get_tiles(f.name, 11, 113)
    assert(rows[113][0]['fields'][3] == 'Lrp1b')

    rows = cdt.get_tiles(f.name, 11, 112)
    assert(rows[112][0]['fields'][3] == 'Lrp1b')


def test_random_importance():
    # check that when aggregating using random importance, all values that
    # are in a higher resolution tile are also in the lower resolution
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data',
                         '25435_PM15-000877_SM-7QK6O.seg')

    result = runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', '2', '--importance-column', 'random',
         '--assembly', 'b37', '--has-header', '--output-file', f.name])

    # import traceback
    a, b, tb = result.exc_info
    '''
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    '''

    cdt.get_tileset_info(f.name)
    # print("tileset_info:", tileset_info)
    # TODO: Make assertions about result

    cdt.get_tiles(f.name, 0, 0)
    # print("rows:", rows)
    # TODO: Make assertions about result

    list(cdt.get_tiles(f.name, 1, 0).values()) + \
        list(cdt.get_tiles(f.name, 1, 1).values())
    # print('rows:', rows)
    # TODO: Make assertions about result

    # check to make sure that tiles in the higher zoom levels
    # are all present in lower zoom levels
    found = {}
    for row in cdt.get_tiles(f.name, 5, 15).values():
        for rect in row:
            found[rect['xStart']] = False

    for row in cdt.get_tiles(f.name, 6, 30).values():
        for rect in row:
            if rect['xStart'] in found:
                found[rect['xStart']] = True

    for row in cdt.get_tiles(f.name, 6, 31).values():
        for rect in row:
            if rect['xStart'] in found:
                found[rect['xStart']] = True

    for key, value in found.items():
        assert(value)

    pass


def test_no_chromosome_limit():
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data',
                         'geneAnnotationsExonsUnions.short.bed')

    result = runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', '60', '--importance-column', '5',
         '--assembly', 'hg19',
         '--output-file', f.name])

    # import traceback
    '''
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    '''
    a, b, tb = result.exc_info

    rows = cdt.get_tiles(f.name, 0, 0)[0]
    foundOther = False

    for row in rows:
        if row['fields'][0] != 'chr1':
            # print("row", row)
            assert(row['xStart'] > 200000000)
        if row['fields'][0] != 'chr14':
            foundOther = True
        break
    # make sure there's chromosome other than 14 in the output
    assert(foundOther)

    os.remove(f.name)
    pass


def test_chromosome_limit():
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data',
                         'geneAnnotationsExonsUnions.short.bed')

    runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', '60', '--importance-column', '5',
         '--assembly', 'hg19', '--chromosome', 'chr14',
         '--output-file', f.name])
    # TODO: Make assertions about result

    # print('output:', result.output, result)
    rows = cdt.get_tiles(f.name, 0, 0)[0]

    for row in rows:
        assert(row['fields'][0] == 'chr14')

    os.remove(f.name)
    pass


def test_float_importance():
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data', 'test_float_importance.bed')

    runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', '2', '--importance-column', '4',
         '--assembly', 'hg38', '--no-header', '--output-file', f.name])
    # TODO: Make assertions about result


"""
def test_get_tiles():
    f = h5py.File('test/sample_data/cnv.hibed')
    data = cht.get_discrete_data(f, 22, 48)

    assert(len(data) > 0)

    data = cht.get_discrete_data(f, 22, 50)
    assert(len(data) > 0)

    data = cht.get_discrete_data(f, 0, 0)
    assert(len(data) == 100)

def check_tile_for_duplicate_entries(discrete_data):
    '''
    Make sure that there are no entries with the same UID in any tile.
    '''
    seen = set()

    for i,d in enumerate(discrete_data):
        uid = d[-2]

        if uid in seen:
            # print("seen uid:", uid)
            # print("d:", d)
            return False

        # print("adding uid:", uid, d[:3])
        seen.add(uid)

    return True


def test_tile_ranges():
    f = h5py.File('test/sample_data/cnv.hibed')

    data11 = cht.get_discrete_data(f, 11, 6)
    assert(check_tile_for_duplicate_entries(data11) == True)

    max_length_11 = max([int(d[2]) - int(d[1]) for d in data11])
    # print("data11:", max_length_11)

    data10 = cht.get_discrete_data(f, 10, 3)
    max_length_10 = max([int(d[2]) - int(d[1]) for d in data10])
    # print("data10:", max_length_10)

    # more zoomed out tiles should have longer tiles than more
    # zoomed in tiles
    assert(max_length_10 >= max_length_11)

    d1 = cht.get_discrete_data(f, 11, 5)
    # print("d1:", len(d1))
    # print("dv:", [x for x in d1 if (int(x[1]) < 12000000
    # and int(x[2]) > 12000000)])

    d3 = cht.get_discrete_data(f, 12, 10)
    # print("d2:", len(d3))

    d4 = cht.get_discrete_data(f, 12, 11)
    # print("d3:", len(d4))

def test_limit_by_chromosome():

"""



import numpy as np


def tileset_info(f, bounds=None):
    if 'min-pos' in f.attrs:
        min_pos = f.attrs['min-pos']
    else:
        min_pos = [0, 0]

    if 'max-pos' in f.attrs:
        max_pos = f.attrs['max-pos']
    else:
        max_pos = f['resolutions']['1']['values'].shape

    return {
        'min_pos': min_pos,
        'max_pos': max_pos,
        'resolutions': [int(r) for r in f['resolutions']],
        'mirror_tiles': 'false',
        'bins_per_dimension': 256,
    }


def tiles(f, z, x, y):
    '''
    Return tiles for the given region.

    Parameters:
    -----------
    f: h5py.File
        File pointer to the hdf5 file containing the matrices
    z: int
        The zoom level
    x: int
        The tile's x position
    y: int
        The tile's y position
    '''
    resolutions = sorted(map(int, f['resolutions'].keys()))[::-1]
    tsinfo = tileset_info(f)
    n_bins = tsinfo['bins_per_dimension']

    if z >= len(resolutions):
        raise ValueError(
            'Zoom level out of bounds:', z,
            "resolutions:", resolutions)

    tile_width = tsinfo['bins_per_dimension']

    # Where in the matrix the tile starts
    tile_x_start = x * tile_width
    tile_y_start = y * tile_width

    tile_x_end = tile_x_start + n_bins
    tile_y_end = tile_y_start + n_bins

    mat = f['resolutions'][str(resolutions[z])]['values']
    data = mat[
        tile_y_start:tile_y_end,
        tile_x_start:tile_x_end]

    x_pad = n_bins - data.shape[0]
    y_pad = n_bins - data.shape[1]

    if x_pad > 0 or y_pad > 0:
        data = np.pad(
            data, ((0, x_pad), (0, y_pad)), 'constant',
            constant_values=(np.nan, np.nan)
        )

    return data



import unittest

import numpy as np
from numpy.testing import assert_array_equal

from clodius.tiles.mrmatrix import tileset_info, tiles


class MockHdf5(dict):
    # By wrapping a dict in our own class, we can add arbitrary attributes.
    pass


class TilesetInfoTest(unittest.TestCase):

    def setUp(self):
        tileset_stub = {
            'resolutions': {
                '1': {
                    'values': np.array([[1, 2], [3, 4]])
                }
            }
        }

        self.tileset = MockHdf5(tileset_stub)
        self.tileset.attrs = {}

        self.tileset_min = MockHdf5(tileset_stub)
        self.tileset_min.attrs = {'min-pos': (1, 1)}

        self.tileset_max = MockHdf5(tileset_stub)
        self.tileset_max.attrs = {'max-pos': (9, 9)}

        self.info = {
            'bins_per_dimension': 256,
            'max_pos': (2, 2),
            'min_pos': [0, 0],
            'mirror_tiles': 'false',
            'resolutions': [1]
        }

    def test_default(self):
        self.assertEqual(tileset_info(self.tileset), self.info)
        self.assertEqual(tileset_info(self.tileset, bounds='???'), self.info)

    def test_with_min(self):
        self.info['min_pos'] = (1, 1)
        self.assertEqual(tileset_info(self.tileset_min), self.info)
        self.assertEqual(tileset_info(
            self.tileset_min, bounds='???'), self.info)

    def test_with_max(self):
        self.info['max_pos'] = (9, 9)
        self.assertEqual(tileset_info(self.tileset_max), self.info)
        self.assertEqual(tileset_info(
            self.tileset_max, bounds='???'), self.info)


class TilesTest(unittest.TestCase):
    def test_zoom_out_of_bounds(self):
        def should_fail():
            tileset_stub = MockHdf5({
                'resolutions': {
                    '1': {
                        'values': np.array([[1, 2], [3, 4]])
                    }
                }
            })
            tileset_stub.attrs = {}
            tiles(tileset_stub, 2, 0, 0)
        self.assertRaisesRegex(
            ValueError, r'Zoom level out of bounds', should_fail)

    def test_padding(self):
        tileset = MockHdf5({
            'resolutions': {
                '1': {
                    'values': np.array([[1.0, 2], [3, 4]])
                    # It's important that there is a float value:
                    # If there isn't, np.nan will be converted
                    # to a large negative integer.
                }
            }
        })
        tileset.attrs = {}
        zoomed = tiles(tileset, 0, 0, 0)
        self.assertEqual(zoomed.shape, (256, 256))
        assert_array_equal(zoomed[0:2, 0:2], [[1, 2], [3, 4]])
        assert_array_equal(zoomed[2:256, 0], [np.nan for x in range(254)])

    def test_bins(self):
        tileset = MockHdf5({
            'resolutions': {
                '1': {
                    'values': np.array(
                        [[float(x) for x in range(500)] for y in range(500)]
                    )
                }
            }
        })
        tileset.attrs = {}

        zoomed_0 = tiles(tileset, 0, 0, 0)
        self.assertEqual(zoomed_0.shape, (256, 256))
        self.assertEqual(zoomed_0[0, 0], 0)

        zoomed_1 = tiles(tileset, 0, 1, 1)
        self.assertEqual(zoomed_1.shape, (256, 256))
        self.assertEqual(zoomed_1[0, 0], 256)
        self.assertEqual(zoomed_1[1, 0], 256)  # Constant dimension
        self.assertEqual(zoomed_1[0, 1], 257)  # Changing dimension
        self.assertEqual(zoomed_1[0, 256 - 13], 499)
        assert_array_equal(
            zoomed_1[0:1, (256 - 12):(256 - 12 + 1)], [[np.nan]])
        # Plain assertEqual gave: nan != nan

    def test_zoom(self):
        tileset = MockHdf5({
            'resolutions': {
                '1': {
                    'values': np.array([[1, 2], [3, 4]])
                },
                '5': {
                    'values': np.array([[3, 4], [5, 6]])
                },
                '11': {
                    'values': np.array([[5, 6], [7, 8]])
                }
            }
        })
        tileset.attrs = {}

        zoomed_0 = tiles(tileset, 0, 0, 0)
        assert_array_equal(zoomed_0[0:2, 0:2], [[5, 6], [7, 8]])

        zoomed_1 = tiles(tileset, 1, 0, 0)
        assert_array_equal(zoomed_1[0:2, 0:2], [[3, 4], [5, 6]])

        zoomed_2 = tiles(tileset, 2, 0, 0)
        assert_array_equal(zoomed_2[0:2, 0:2], [[1, 2], [3, 4]])



import unittest
from tempfile import TemporaryDirectory
import csv
from math import nan

import numpy as np
from numpy.testing import assert_array_equal
import h5py

from scripts.tsv_to_mrmatrix import coarsen, parse, get_height, get_width


class CoarsenTest(unittest.TestCase):
    def test_5_layer_pyramid(self):
        tile_size = 4
        max_zoom = 4
        max_width = tile_size * 2 ** max_zoom

        with TemporaryDirectory() as tmp_dir:
            hdf5 = h5py.File(tmp_dir + '/temp.hdf5', 'w')
            g = hdf5.create_group('resolutions')
            g1 = g.create_group('1')
            ds = g1.create_dataset('values', (max_width, max_width),
                                   dtype='f4', compression='lzf',
                                   fillvalue=np.nan)
            for y in range(max_width):
                a = np.array([float(x) for x in range(max_width)])
                ds[y, :max_width] = a

            # before coarsen()
            self.assertEqual(list(hdf5.keys()), ['resolutions'])
            self.assertEqual(list(hdf5['resolutions'].keys()), ['1'])
            self.assertEqual(list(hdf5['resolutions']['1'].keys()), ['values'])
            self.assertEqual(
                list(hdf5['resolutions']['1']['values'].shape), [64, 64])
            self.assertEqual(
                hdf5['resolutions']['1']['values'][:].tolist()[0],
                [float(x) for x in range(64)]
            )

            coarsen(hdf5, type='values', tile_size=tile_size)

            # after coarsen()
            self.assertEqual(list(hdf5.keys()), ['resolutions'])
            self.assertEqual(list(hdf5['resolutions'].keys()), [
                             '1', '16', '2', '4', '8'])
            self.assertEqual(
                list(hdf5['resolutions']['16'].keys()), ['values'])
            shapes = {
                '1': 64,
                '2': 32,
                '4': 16,
                '8': 8,
                '16': 4
            }
            for (k, v) in shapes.items():
                self.assertEqual(hdf5['resolutions'][k]
                                 ['values'].shape, (v, v))
            row = [1920, 6016, 10112, 14208]
            self.assertEqual(
                hdf5['resolutions']['16']['values'][:].tolist(),
                [row, row, row, row])

    def test_math(self):
        tile_size = 2
        max_zoom = 2
        max_width = tile_size * 2 ** max_zoom

        with TemporaryDirectory() as tmp_dir:
            hdf5 = h5py.File(tmp_dir + '/temp.hdf5', 'w')
            g = hdf5.create_group('resolutions')
            g1 = g.create_group('1')
            ds = g1.create_dataset('values', (max_width, max_width),
                                   dtype='f4', compression='lzf',
                                   fillvalue=np.nan)
            for y in range(max_width):
                a = np.array([float(x) for x in range(max_width)])
                ds[y, :max_width] = a

            coarsen(hdf5, type='values', tile_size=tile_size)

            # after coarsen()
            self.assertEqual(list(hdf5.keys()), ['resolutions'])
            self.assertEqual(list(hdf5['resolutions'].keys()), ['1', '2', '4'])

            shapes = {
                '1': 8,
                '2': 4,
                '4': 2
            }
            for (k, v) in shapes.items():
                self.assertEqual(hdf5['resolutions'][k]
                                 ['values'].shape, (v, v))

            row8 = list(range(8))
            assert_array_equal(
                hdf5['resolutions']['1']['values'],
                [row8 for _ in range(8)])

            row4 = [8 * x + 2 for x in range(4)]
            assert_array_equal(
                hdf5['resolutions']['2']['values'],
                [row4 for _ in range(4)])

            row2 = [24, 88]
            assert_array_equal(
                hdf5['resolutions']['4']['values'],
                [row2 for _ in range(2)])


class ParseTest(unittest.TestCase):
    def test_wide_labelled_square(self):
        with TemporaryDirectory() as tmp_dir:
            csv_path = tmp_dir + '/tmp.csv'
            with open(csv_path, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter='\t')
                # header:
                col_labels = ['col-{}'.format(x) for x in range(513)]
                writer.writerow(col_labels)
                # body:
                for y in range(0, 3):
                    writer.writerow(['row-{}'.format(y)] + [0] * 512)
                for y in range(3, 6):
                    writer.writerow(['row-{}'.format(y)] + [1] * 512)
                for y in range(6, 9):
                    writer.writerow(['row-{}'.format(y)] + [1, -1] * 256)
            csv_handle = open(csv_path, 'r')

            hdf5_path = tmp_dir + 'tmp.hdf5'
            hdf5_write_handle = h5py.File(hdf5_path, 'w')

            height = get_height(csv_path)
            width = get_width(csv_path, is_labelled=True)
            parse(csv_handle, hdf5_write_handle, height, width,
                  delimiter='\t', first_n=None, is_labelled=True)

            hdf5 = h5py.File(hdf5_path, 'r')
            self.assertEqual(list(hdf5.keys()),
                             ['col_labels', 'resolutions', 'row_labels'])
            self.assertEqual(list(hdf5['col_labels']),
                             col_labels[1:])
            self.assertEqual(list(hdf5['row_labels']),
                             ['row-{}'.format(r) for r in range(9)])

            self.assertEqual(list(hdf5['resolutions'].keys()),
                             ['1', '2', '2-nan_values'])

            self.assertEqual(list(hdf5['resolutions']['1'].keys()), [
                             'nan_values', 'values'])
            assert_array_equal(
                hdf5['resolutions']['1']['nan_values'], [[0] * 512] * 512
            )
            res_1 = hdf5['resolutions']['1']['values']
            assert_array_equal(res_1[0], [0] * 512)
            assert_array_equal(res_1[3], [1] * 512)
            assert_array_equal(res_1[6], [1, -1] * 256)
            assert_array_equal(res_1[9], [nan] * 512)

            self.assertEqual(list(hdf5['resolutions']['2'].keys()), ['values'])
            # TODO: We are missing nan_values at higher aggregations: Bug?
            # https://github.com/higlass/clodius/issues/62
            res_2 = hdf5['resolutions']['2']['values']
            assert_array_equal(res_2[0], [0] * 256)
            # Stradles the 0 and 1 rows
            assert_array_equal(res_2[1], [2] * 256)
            assert_array_equal(res_2[2], [4] * 256)
            assert_array_equal(res_2[3], [0] * 256)  # -1 and +1 cancel out
            assert_array_equal(res_2[4], [0] * 256)
            assert_array_equal(res_2[5], [0] * 256)
            assert_array_equal(res_2[6], [0] * 256)
            # TODO: We lose nan at higher aggregations.
            # https://github.com/higlass/clodius/issues/62

    def _assert_unlabelled_roundtrip_lt_256(
            self, matrix, delimiter):
        with TemporaryDirectory() as tmp_dir:
            csv_path = tmp_dir + '/tmp.csv'
            with open(csv_path, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=delimiter)
                # body:
                for row in matrix:
                    writer.writerow(row)

            csv_handle = open(csv_path, 'r')
            hdf5_path = tmp_dir + 'tmp.hdf5'
            hdf5_write_handle = h5py.File(hdf5_path, 'w')

            is_labelled = False
            height = get_height(csv_path, is_labelled=is_labelled)
            width = get_width(csv_path, is_labelled=is_labelled)
            parse(csv_handle, hdf5_write_handle, height, width,
                  first_n=None, is_labelled=is_labelled,
                  delimiter=delimiter)

            hdf5 = h5py.File(hdf5_path, 'r')
            self.assertEqual(list(hdf5.keys()), ['resolutions'])
            self.assertEqual(list(hdf5['resolutions'].keys()), ['1'])
            self.assertEqual(list(hdf5['resolutions']['1'].keys()),
                             ['nan_values', 'values'])
            assert_array_equal(
                hdf5['resolutions']['1']['nan_values'],
                [[0] * len(matrix[0])] * len(matrix)
            )
            assert_array_equal(
                hdf5['resolutions']['1']['values'],
                matrix
            )

    def test_unlabelled_csv(self):
        self._assert_unlabelled_roundtrip_lt_256(
            matrix=[[x + y for x in range(4)] for y in range(4)],
            delimiter=','
        )

    def _assert_unlabelled_roundtrip_1024(
            self, matrix, first_row=None, first_col=None, first_n=None):
        delimiter = '\t'
        with TemporaryDirectory() as tmp_dir:
            csv_path = tmp_dir + '/tmp.csv'
            with open(csv_path, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=delimiter)
                # body:
                for row in matrix:
                    writer.writerow(row)

            csv_handle = open(csv_path, 'r')
            hdf5_path = tmp_dir + 'tmp.hdf5'
            hdf5_write_handle = h5py.File(hdf5_path, 'w')

            is_labelled = False
            height = get_height(csv_path, is_labelled=is_labelled)
            width = get_width(csv_path, is_labelled=is_labelled)
            parse(csv_handle, hdf5_write_handle, height, width,
                  first_n=first_n, is_labelled=is_labelled,
                  delimiter=delimiter)

            hdf5 = h5py.File(hdf5_path, 'r')
            self.assertEqual(list(hdf5.keys()), ['resolutions'])
            self.assertEqual(list(hdf5['resolutions'].keys()),
                             ['1', '2', '2-nan_values', '4', '4-nan_values'])
            self.assertEqual(list(hdf5['resolutions']['1'].keys()),
                             ['nan_values', 'values'])
            self.assertEqual(list(hdf5['resolutions']['4'].keys()),
                             ['values'])
            res_4 = hdf5['resolutions']['4']['values']
            if first_row:
                assert_array_equal(res_4[0], first_row)
            if first_col:
                assert_array_equal(
                    [res_4[y][0] for y in range(len(first_col))],
                    first_col)

    def test_unlabelled_tsv_tall(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(4)] for y in range(1000)],
            first_col=[16] * 250 + [0] * 6
        )

    def test_unlabelled_tsv_wide(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(1000)] for y in range(4)],
            first_row=[16] * 250 + [0] * 6
        )

    def test_unlabelled_tsv_tall_first_n(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(4)] for y in range(1000)],
            first_col=[8] + [0] * 255,
            first_n=2
        )

    def test_unlabelled_tsv_wide_first_n(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(1000)] for y in range(4)],
            first_row=[8] * 250 + [0] * 6,
            first_n=2
        )



from __future__ import print_function

import clodius.db_tiles as cdt
import clodius.hdf_tiles as cht
import click.testing as clt
import clodius.cli.aggregate as cca
import h5py
import negspy.coordinates as nc
import numpy as np
import os.path as op
import sys

sys.path.append("scripts")

testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedfile():
    input_file = op.join(testdir, 'sample_data',
                         'geneAnnotationsExonsUnions.hg19.short.bed')
    output_file = '/tmp/geneAnnotationsExonsUnions.hg19.short.bed'

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', 20,
         '--importance-column', 5,
         '--assembly', 'hg19',
         '--output-file', output_file,
         '--delimiter', '\t'])

    a, b, tb = result.exc_info
    '''
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    '''
    assert(result.exit_code == 0)

    results = cdt.get_tiles(output_file, 6, 3, num_tiles=1)
    # print("results:", results)

    assert(len(results[3][0]['fields']) == 14)

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedfile,
        [input_file,
         '--max-per-tile', 20,
         '--importance-column', 5,
         '--assembly', 'hg19',
         '--output-file', output_file])

    assert(result.exit_code == 0)

    results = cdt.get_tiles(output_file, 6, 3, num_tiles=3)

    assert(len(results[3][0]['fields']) == 17)


testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedgraph():
    input_file = op.join(testdir, 'sample_data', 'cnvs_hw.tsv')
    assembly_file = op.join(testdir, 'sample_data', 'test_cnvs_assembly')
    output_file = '/tmp/cnvs_hw.hitile'

    # run once to make sure it doesn't crash on a smaller genome
    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [input_file,
         '--output-file', output_file,
         # '--assembly', 'grch37',
         '--chromsizes-filename', assembly_file,
         '--chromosome-col', '2',
         '--from-pos-col', '3',
         '--to-pos-col', '4',
         '--value-col', '5',
         '--has-header',
         '--nan-value', 'NA'])

    # run again with the proper assembly
    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [input_file,
         '--output-file', output_file,
         '--assembly', 'grch37',
         # '--chromsizes-filename', assembly_file,
         '--chromosome-col', '2',
         '--from-pos-col', '3',
         '--to-pos-col', '4',
         '--value-col', '5',
         '--has-header',
         '--nan-value', 'NA'])

    '''
    import traceback
    a,b,tb = result.exc_info
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    '''

    assert(result.exit_code == 0)
    f = h5py.File(output_file)
    # print("tile_0_0", d)

    # print("tile:", cht.get_data(f, 22, 0))
    # return
    d = cht.get_data(f, 0, 0)

    assert(not np.isnan(d[0]))
    assert(np.isnan(d[-1]))
    cht.get_data(f, 3, 0)
    # TODO: Make assertions about result

    # print("prev_tile_3_0:", prev_tile_3_0)

    assert(result.exit_code == 0)

    # TODO: Why are we ignoring these?
    # assert(sum(prev_tile_3_0) < 0)
    # input_file = op.join(testdir, 'sample_data', 'cnvs_hw.tsv.gz')
    # result = runner.invoke(
    # cca.bedgraph,
    # [input_file,
    # '--output-file', output_file,
    # '--assembly', 'grch37',
    # '--chromosome-col', '2',
    # '--from-pos-col', '3',
    # '--to-pos-col', '4',
    # '--value-col', '5',
    # '--has-header',
    # '--nan-value', 'NA'])
    # '''
    # import traceback
    # print("exc_info:", result.exc_info)
    # a,b,tb = result.exc_info
    # print("result:", result)
    # print("result.output", result.output)
    # print("result.error", traceback.print_tb(tb))
    # print("Exception:", a,b)
    # '''
    # f = h5py.File(output_file)
    # tile_3_0 = cht.get_data(f, 3, 0)
    # assert(sum(tile_3_0) - sum(prev_tile_3_0) < 0.0001)


testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedpe():
    input_file = op.join(testdir, 'sample_data',
                         'Rao_RepA_GM12878_Arrowhead.txt')
    output_file = '/tmp/bedpe.db'

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedpe,
        [input_file,
         '--output-file', output_file,
         '--assembly', 'hg19',
         '--chr1-col', '1',
         '--from1-col', '2',
         '--to1-col', '3',
         '--chr2-col', '1',
         '--from2-col', '2',
         '--to2-col', '3'])

    '''
    print("result:", result)
    print("result.output", result.output)
    '''

    assert(result.exit_code == 0)

    tiles = cdt.get_2d_tiles(output_file, 0, 0, 0, numx=1, numy=1)

    assert('\n' not in tiles[(0, 0)][0]['fields'][2])


testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedgraph1():
    input_file = op.join(testdir, 'sample_data', 'dm3_values.tsv')
    output_file = '/tmp/dm3_values.hitile'

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [input_file,
         '--output-file', output_file,
         '--assembly', 'dm3'])

    a, b, tb = result.exc_info

    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """

    # print("result.output", result.output)

    f = h5py.File('/tmp/dm3_values.hitile')
    # max_zoom = f['meta'].attrs['max-zoom']
    # TODO: Make assertions about result
    values = f['values_0']

    import numpy as np
    # print("values:", values[8])
    # genome positions are 0 based as stored in hitile files
    assert(np.isnan(values[8]))
    assert(values[9] == 1)
    assert(values[10] == 1)
    assert(values[13] == 1)
    assert(np.isnan(values[14]))
    assert(np.isnan(values[15]))

    chrom_info = nc.get_chrominfo('dm3')
    chr_2r_pos = nc.chr_pos_to_genome_pos('chr2R', 0, chrom_info)
    # print('chr_2r_pos:', chr_2r_pos)

    assert(np.isnan(values[chr_2r_pos + 28]))
    assert(values[chr_2r_pos + 29] == 77)
    assert(values[chr_2r_pos + 38] == 77)
    assert(values[chr_2r_pos + 39] == 0)

    assert(result.exit_code == 0)

    d = cht.get_data(f, 0, 0)
    # print("d[:10]", d[:10])
    # print("sum(d):", sum([x for x in d if not np.isnan(x)]))
    assert(np.nansum(d) > 1.0 and np.nansum(d) < 10.0)

    return

    input_file = op.join(testdir, 'sample_data', 'test3chroms_values.tsv')
    output_file = '/tmp/test3chroms_values.hitile'

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [input_file,
         '--output-file', output_file,
         '--assembly', 'test3chroms'])

    # print('output:', result.output, result)

    f = h5py.File('/tmp/test3chroms_values.hitile')
    # f['meta'].attrs['max-zoom']
    # TODO: Make assertions about result

    # print('max_zoom:', max_zoom)
    # print("len", len(f['values_0']))

    values = f['values_0']

    # print('values', values[:100])

    # genome positions are 0 based as stored in hitile files
    assert(values[8] == 0)
    assert(values[9] == 1)
    assert(values[10] == 1)
    assert(values[13] == 1)
    assert(values[14] == 0)
    assert(values[15] == 0)

    chr2_pos = nc.chr_pos_to_genome_pos('chr2', 0, 'test3chroms')

    assert(values[chr2_pos + 28] == 0)
    assert(values[chr2_pos + 29] == 77)
    assert(values[chr2_pos + 38] == 77)
    assert(values[chr2_pos + 39] == 0)

    assert(result.exit_code == 0)

    d = cht.get_data(f, 0, 0)
    assert(sum(d) == 770 + 880 + 5)
    # print("d:", d)


testdir = op.realpath(op.dirname(__file__))



import clodius.tiles.nplabels as ctn
import h5py


def test_tiles():
    filename = 'data/labels.h5'
    f = h5py.File(filename, 'r')

    array = f['labels']

    tiles = ctn.tiles(array, 0, 0, None)

    assert(tiles[0]['x'] == 0)
    assert(tiles[-1]['x'] == 86992)

    tiles = ctn.tiles(array, 2, 1, None)

    assert(tiles[0]['x'] == 32768)
    assert(tiles[-1]['x'] == 65535)



import functools as ft
import clodius.tiles.density as hgde
import clodius.tiles.utils as hgut


def test_bundled_tiles():
    filename = 'data/points_density.h5'

    tiles = hgut.bundled_tiles_wrapper_2d(
        ['c.2.0.0', 'c.2.1.1', 'c.2.0.1', 'c.2.1.0'],
        ft.partial(hgde.tiles, filename))
    assert(len(tiles) == 4)



from __future__ import print_function

import click.testing as clt
import clodius.cli.convert as ccc
import clodius.tiles.multivec as ctv
import os.path as op
import tempfile

testdir = op.realpath(op.dirname(__file__))


def test_bedfile_to_multivec():
    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data', 'sample.multival.bed')
    chromsizes_file = op.join(testdir, 'sample_data', 'sample.multival.chrom.sizes')

    with tempfile.TemporaryDirectory() as tmp_dir:
        out_file = op.join(tmp_dir, 'out.multivec')

        _ = runner.invoke(
            ccc.bedfile_to_multivec,
            [input_file,
             '--output-file', out_file,
             '--assembly', 'hg38',
             '--num-rows', 3,
             '--chromsizes-filename', chromsizes_file,
             '--starting-resolution', '1000'])

        # import traceback
        # a, b, tb = result.exc_info

        # print("exc_info:", result.exc_info)
        # print("result:", result)
        # print("result.output", result.output)
        # print("result.error", traceback.print_tb(tb))
        # print("Exception:", a,b)

        tsinfo = ctv.tileset_info(out_file)
        # print("tsinfo:", tsinfo)

        assert 'resolutions' in tsinfo
        assert tsinfo['max_pos'][0] == 18000
        tile = (ctv.get_single_tile(out_file, (0, 0)))

        # input_file:
        # chr1    0   1000    1.0 2.0 3.0
        # chr1    1000    2000
        # chr2    5000    6000    20.0    30.0    40.0
        #
        # # input chromsizes
        # chr1  10000
        # chr2    8000

        # first row, first chrom first value
        assert len(tile) == 3

        assert tile[0][0] == 1.0
        assert tile[0][15] == 20.0

        assert tile[1][0] == 2.0
        assert tile[2][0] == 3.0
        assert tile[2][15] == 40.0


def test_load_multivec_tiles():
    op.join(testdir, 'sample_data', 'sample.bed.multires.mv5')
    # TODO: Make assertions about result


def test_states_format_befile_to_multivec():
    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data',
                         'states_format_input_testfile.bed.gz')
    rows_info_file = op.join(testdir, 'sample_data',
                             'states_format_test_row_infos.txt')
    tempfile.NamedTemporaryFile(delete=False)
    # TODO: Make assertions about result
    # print("input_file", input_file)

    result = runner.invoke(
        ccc.bedfile_to_multivec,
        [input_file,
         '--format', 'states',
         '--row-infos-filename', rows_info_file,
         '--assembly', 'hg38',
         '--starting-resolution', '200',
         '--num-rows', '10'])

    # import traceback
    a, b, tb = result.exc_info
    '''
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    '''


def test_ignore_bedfile_headers():
    runner = clt.CliRunner()
    input_file = op.join(testdir, 'sample_data',
                         '3_header_100_testfile.bed.gz')
    rows_info_file = op.join(testdir, 'sample_data',
                             '3_header_100_row_infos.txt')
    tempfile.NamedTemporaryFile(delete=False)
    # TODO: Make assertions about result

    result = runner.invoke(
        ccc.bedfile_to_multivec,
        [input_file,
         '--format', 'states',
         '--row-infos-filename', rows_info_file,
         '--assembly', 'hg19',
         '--starting-resolution', '200',
         '--num-rows', '15'])

    # import traceback
    a, b, tb = result.exc_info



from __future__ import division, print_function
import json

import numpy as np
import pandas as pd
import cooler
import h5py
import time

TILESIZE = 256

where = np.flatnonzero
chromsizes = cooler.read_chromsizes(
    'http://s3.amazonaws.com/pkerp/data/mm9/chromInfo.txt')  # defaults to reading chr#,X,Y,M
chromosomes = list(chromsizes.keys())
chromid_map = dict(zip(chromosomes, range(len(chromosomes))))
cumul_lengths = np.r_[0, np.cumsum(chromsizes)]


def absCoord2bin(c, pos):
    try:
        cid = where(cumul_lengths > pos)[0] - 1
    except IndexError:
        return c.info['nbins']
    chrom = chromosomes[cid]
    relPos = pos - cumul_lengths[cid]
    return c.offset((chrom, relPos, chromsizes[chrom]))


def getData(FILEPATH, zoomLevel, startPos1, endPos1, startPos2, endPos2):

    groupname = str(zoomLevel)

    with h5py.File(FILEPATH, 'r') as f:
        c = cooler.Cooler(f[groupname])

        i0 = absCoord2bin(c, startPos1)
        i1 = absCoord2bin(c, endPos1)
        j0 = absCoord2bin(c, startPos2)
        j1 = absCoord2bin(c, endPos2)
        mat = c.matrix(balance=True)[i0:i1, j0:j1]

    flat = list(mat.toarray().ravel())

    return json.dumps({'dense': flat})


def getData2(cooler_matrix, zoomLevel, startPos1, endPos1, startPos2, endPos2):

    c = cooler_matrix['cooler']
    matrix = cooler_matrix['matrix']

    i0 = absCoord2bin(c, startPos1)
    i1 = absCoord2bin(c, endPos1)
    j0 = absCoord2bin(c, startPos2)
    j1 = absCoord2bin(c, endPos2)

    if (i1 - i0) == 0 or (j1 - j0) == 0:
        return pd.DataFrame(columns=['genome_start', 'genome_end', 'balanced'])

    t1 = time.time()
    pixels = matrix[i0:i1, j0:j1]
    cid1 = pixels['chrom1'].apply(chromid_map.get)
    cid2 = pixels['chrom2'].apply(chromid_map.get)
    pixels['genome_start'] = cumul_lengths[cid1] + pixels['start1']
    pixels['genome_end'] = cumul_lengths[cid2] + pixels['end2']

    print("z:", zoomLevel, "t:", time.time() - t1)
    return pixels[['genome_start', 'genome_end', 'balanced']]


def getData3(cooler_matrix, zoomLevel, startPos1, endPos1, startPos2, endPos2):
    c = cooler_matrix['cooler']

    i0 = absCoord2bin(c, startPos1)
    i1 = absCoord2bin(c, endPos1)
    j0 = absCoord2bin(c, startPos2)
    j1 = absCoord2bin(c, endPos2)

    if (i1 - i0) == 0 or (j1 - j0) == 0:
        return pd.DataFrame(columns=['genome_start', 'genome_end', 'balanced'])

    pixels = c.matrix(as_pixels=True, max_chunk=np.inf)[i0:i1, j0:j1]

    if not len(pixels):
        return pd.DataFrame(columns=['genome_start', 'genome_end', 'balanced'])

    lo = min(i0, j0)
    hi = max(i1, j1)
    bins = c.bins()[['chrom', 'start', 'end', 'weight']][lo:hi]
    bins['chrom'] = bins['chrom'].cat.codes
    pixels = cooler.annotate(pixels, bins)
    pixels['genome_start'] = cumul_lengths[pixels['chrom1']] + pixels['start1']
    pixels['genome_end'] = cumul_lengths[pixels['chrom2']] + pixels['end2']
    pixels['balanced'] = pixels['count'] * \
        pixels['weight1'] * pixels['weight2']

    return pixels[['genome_start', 'genome_end', 'balanced']]


def getInfo(FILEPATH):

    with h5py.File(FILEPATH, 'r') as f:
        total_length = int(cumul_lengths[-1])
        binsize = int(f['0'].attrs['bin-size'])
        binsize = 1000
        n_tiles = total_length / binsize / TILESIZE
        print("total_length:", total_length, binsize, TILESIZE)
        n_zooms = int(np.ceil(np.log2(n_tiles)))
        max_width = binsize * TILESIZE * 2**n_zooms

        info = {
            'min_pos': [0.0, 0.0],
            'max_pos': [total_length, total_length],
            'max_zoom': n_zooms,
            'max_width': max_width,
            'bins_per_dimension': TILESIZE,
        }

    return info

    return json.dumps(info)



from setuptools import setup
import os
import io


HERE = os.path.dirname(os.path.abspath(__file__))


def read(*parts, **kwargs):
    filepath = os.path.join(HERE, *parts)
    encoding = kwargs.pop('encoding', 'utf-8')
    with io.open(filepath, encoding=encoding) as fh:
        text = fh.read()
    return text


def get_requirements(path):
    content = read(path)
    return [
        req
        for req in content.split("\n")
        if req != '' and not req.startswith('#')
    ]


setup_requires = [
    'numpy',
]

install_requires = get_requirements('requirements.txt')

setup(
    name='clodius',
    version='0.10.13',
    description='Tile generation for big data',
    author='Peter Kerpedjiev',
    author_email='pkerpedjiev@gmail.com',
    url='',
    packages=['clodius', 'clodius.cli', 'clodius.tiles'],
    setup_requires=setup_requires,
    install_requires=install_requires,
    scripts=[
        'scripts/tsv_to_mrmatrix.py'
    ],
    entry_points={
        'console_scripts': [
            'clodius = clodius.cli.aggregate:cli'
        ]
    }
)



import os.path as op
import clodius.tiles.multivec as hgmu


def test_multivec():
    filename = op.join('data', 'all.KL.bed.multires.mv5')

    hgmu.tileset_info(filename)
    # TODO: Make assertions about info returned.



import clodius.tiles.beddb as hgbe
import os.path as op


def test_get_tiles():
    filename = op.join('data', 'corrected.geneListwithStrand.bed.multires')

    hgbe.tiles(filename, ['x.1.0', 'x.1.1'])
    # TODO: Do something with the return value


def test_list_items():
    filename = op.join('data', 'gene_annotations.short.db')

    hgbe.list_items(filename, 0, 100000000, max_entries=100)
    # TODO: Do something with the return value
